@@ -525,7 +525,10 @@ void build_impute_node(ImputeNode &imputer,    WorkerMemory &workspace,
                     {
                         for (int cat = 0; cat < input_data.ncat[col]; cat++)
                         {
-                            imputer.cat_sum[col][cat] += imputer_tree[curr_tree].cat_sum[col][cat] / imputer.cat_weight[col];
+                            imputer.cat_sum[col][cat]
+                                +=
+                            (imputer_tree[curr_tree].cat_sum[col][cat] > 0)?
+                                (imputer_tree[curr_tree].cat_sum[col][cat] / imputer.cat_weight[col]) : 0.;
                             imputer.cat_weight[col]    =  wsum / (double)(2 * look_aboves);
                         }
                         break;
@@ -1207,12 +1210,15 @@ void check_for_missing(InputData &input_data,
         #pragma omp parallel for schedule(static) num_threads(nthreads) shared(input_data)
         for (size_t_for row = 0; row < input_data.nrows; row++)
         {
-            for (size_t col = 0; col < input_data.ncols_numeric; col++)
+            if (input_data.Xc_indptr == NULL)
             {
-                if (is_na_or_inf(input_data.numeric_data[row + col * input_data.nrows]))
+                for (size_t col = 0; col < input_data.ncols_numeric; col++)
                 {
-                    input_data.has_missing[row] = true;
-                    break;
+                    if (is_na_or_inf(input_data.numeric_data[row + col * input_data.nrows]))
+                    {
+                        input_data.has_missing[row] = true;
+                        break;
+                    }
                 }
             }
 
