@@ -1058,6 +1058,8 @@ void divide_subset_split(size_t ix_arr[], size_t st, size_t end, size_t col_num,
 {
     /* TODO: this is a mess, needs refactoring */
     /* TODO: when moving zeros, would be better to instead move by '>' (opposite as in here) */
+    /* TODO: should create an extra version to go along with 'predict' that would
+       add the range penalty right here to spare operations. */
     if (Xc_indptr[col_num] == Xc_indptr[col_num + 1])
     {
         if (missing_action == Fail)
@@ -1253,7 +1255,7 @@ void divide_subset_split(size_t ix_arr[], size_t st, size_t end, size_t col_num,
                 if (Xc_ind[curr_pos] == *row)
                 {
                     if (isnan(Xc[curr_pos])) has_NAs = true;
-                    if (Xc[curr_pos] <= split_point && !isnan(Xc[curr_pos]))
+                    if (!isnan(Xc[curr_pos]) && Xc[curr_pos] <= split_point)
                     {
                         temp       = ix_arr[st];
                         ix_arr[st] = *row;
@@ -1896,52 +1898,6 @@ void todense(size_t ix_arr[], size_t st, size_t end,
     }
 }
 
-template <class sparse_ix>
-bool check_indices_are_sorted(sparse_ix indices[], size_t n)
-{
-    if (n <= 1)
-        return true;
-    if (indices[n-1] < indices[0])
-        return false;
-    for (size_t ix = 1; ix < n; ix++)
-        if (indices[ix] < indices[ix-1])
-            return false;
-    return true;
-}
-
-template <class real_t, class sparse_ix>
-void sort_csc_indices(real_t *restrict Xc, sparse_ix *restrict Xc_ind, sparse_ix *restrict Xc_indptr, size_t ncols_numeric)
-{
-    std::vector<double> buffer_sorted_vals;
-    std::vector<sparse_ix> buffer_sorted_ix;
-    std::vector<size_t> argsorted;
-    size_t n_this;
-    size_t ix1, ix2;
-    for (size_t col = 0; col < ncols_numeric; col++)
-    {
-        ix1 = Xc_indptr[col];
-        ix2 = Xc_indptr[col+1];
-        n_this = ix2 - ix1;
-        if (n_this && !check_indices_are_sorted(Xc_ind + ix1, n_this))
-        {
-            if (buffer_sorted_vals.size() < n_this)
-            {
-                buffer_sorted_vals.resize(n_this);
-                buffer_sorted_ix.resize(n_this);
-                argsorted.resize(n_this);
-            }
-            std::iota(argsorted.begin(), argsorted.begin() + n_this, ix1);
-            std::sort(argsorted.begin(), argsorted.begin() + n_this,
-                      [&Xc_ind](const size_t a, const size_t b){return Xc_ind[a] < Xc_ind[b];});
-            for (size_t ix = 0; ix < n_this; ix++)
-                buffer_sorted_ix[ix] = Xc_ind[argsorted[ix]];
-            std::copy(buffer_sorted_ix.begin(), buffer_sorted_ix.begin() + n_this, Xc_ind + ix1);
-            for (size_t ix = 0; ix < n_this; ix++)
-                buffer_sorted_vals[ix] = Xc[argsorted[ix]];
-            std::copy(buffer_sorted_vals.begin(), buffer_sorted_vals.begin() + n_this, Xc + ix1);
-        }
-    }
-}
 
 bool interrupt_switch = false;
 bool handle_is_locked = false;
